#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import subprocess
import warnings
import os
import glob
from datetime import datetime

import multiprocessing
import numpy as np
import json

from obspy import read_events, UTCDateTime
from obspy.taup import TauPyModel
from obspy.geodetics import locations2degrees, gps2dist_azimuth

from pylot.core.util.utils import identifyPhaseID
from pylot.core.util.obspyDMT_interface import qml_from_obspyDMT
from pylot.tomography.utils import pol2cart, pol2cart_vector
from pylot.tomography.utils import get_metadata


class Propgrid(object):
    '''
    small class that is built from an fm3d propgrid.in file; generates a regular grid in the same way as fm3d
    '''
    def __init__(self, filename):
        self.r_earth = 6371. # earth radius in km
        self.init_propgrid(filename)

    def init_propgrid(self, filename_propgrid):
        self.read_propgrid(filename_propgrid)
        self.init_rGrid()
        self.init_latGrid()
        self.init_longGrid()

    def init_rGrid(self):
        self.rs = np.zeros(self.nR)
        self.rbot = self.r_earth + self.zTop - (self.nR - 1) * self.deltaR
        for index in range(self.nR):
            self.rs[index] = self.rbot + index * self.deltaR

    def init_latGrid(self):
        self.lats = np.zeros(self.nLat)
        self.latS = np.deg2rad(self.latSdeg)
        self.deltaLat = np.deg2rad(self.deltaLatDeg)
        for index in range(self.nLat):
            self.lats[index] = self.latS + index * self.deltaLat

    def init_longGrid(self):
        self.longs = np.zeros(self.nLong)
        self.longW = np.deg2rad(self.longWdeg)
        self.deltaLong = np.deg2rad(self.deltaLongDeg)
        for index in range(self.nLong):
            self.longs[index] = self.longW + index * self.deltaLong

    def read_propgrid(self, filename_propgrid):
        infile = open(filename_propgrid, 'r')
        self.nR, self.nLat, self.nLong = [int(value) for value in infile.readline().split()]
        self.deltaR, self.deltaLatDeg, self.deltaLongDeg = [np.float64(value) for value in infile.readline().split()]
        self.zTop, self.latSdeg, self.longWdeg = [np.float64(value) for value in infile.readline().split()]
        infile.close()


def check_event_notes(eventdir):
    eventfile = os.path.join(eventdir, 'notes.txt')
    if os.path.isfile(eventfile):
        with open(eventfile, 'r') as infile:
            notes = infile.readline()
            print(notes)
            if 'exclude' in notes:
                print('Will exclude this event!')
                return False
    else:
        print('No notes file found.')
    return True


def prepare_fmtomo_dir_first_run(fmtomo_dir):
    """
    This helper function calls the file 'fm3d_prepare_tele' in the fmtomo binary after creating a "fake" receivers.in
    and sources.in file so that fm3d can prepare the teleseismic run and writes the file "init_nodes.out" that contains
    information on the boundary nodes of the current propagation grid. Make sure that "grid3dg" has already been called
    for this function to work.
    :param fmtomo_dir:
    :return:
    """
    def write1(fid):
        fid.write('1\n')

    def write_rec_file():
        """ Writes a fake receiver (has to be inside the grid) """
        with open(recfile, 'w') as fid:
            write1(fid)
            fid.write(f'0    {clat}    {clon}\n')
            for _ in range(3):
                write1(fid)

    def write_src_file():
        """ Writes a fake source """
        fakesrc_str = '''
            1
            1
            P
            0.0000    0.0000   0.0000     0.00000     0.00000     0.00000
            1
            1
            2           1
            1
            1
            '''

        with open(srcfile, 'w') as fid:
            fid.write(fakesrc_str)

    def get_clat_clon():
        if not os.path.isfile(propgrid_file):
            raise IOError(f'Missing file {propgrid_file} for fmtomo first run preparation.')

        with open(propgrid_file, 'r') as fid:
            _, nlat, nlon = (int(item) for item in fid.readline().split())
            _, dlat, dlon = (float(item) for item in fid.readline().split())
            _, lat0, lon0 = (float(item) for item in fid.readline().split())
        clat = lat0 + nlat / 2 * dlat
        clon = lon0 + nlon / 2 * dlon
        return clat, clon

    # check if binary file exists
    fmtomo_prep_tele = os.path.join(fmtomo_dir, 'fm3d_prepare_tele')
    assert os.path.isfile(fmtomo_prep_tele), f'Missing binary file {fmtomo_prep_tele}'

    # set filenames for propgrid, sources and receivers
    propgrid_file = os.path.join(fmtomo_dir, 'propgrid.in')
    recfile = os.path.join(fmtomo_dir, 'receivers.in')
    srcfile = os.path.join(fmtomo_dir, 'sources.in')

    # get coords from propgrid
    clat, clon = get_clat_clon()

    # write fake source and receiver files
    write_src_file()
    write_rec_file()

    # execute fm3d_prepare tele from local directory
    curdir = os.getcwd()
    os.chdir(fmtomo_dir)
    rval = subprocess.check_output([fmtomo_prep_tele]).decode('utf-8')
    if not rval.split('\n')[-2] == ' Finished teleseismic preparation. Stop.':
        raise ValueError('Unexpected output initialisation run.')

    if not os.path.isfile(os.path.join(fmtomo_dir, 'init_nodes.out')):
        raise Exception('Could not create output file init_nodes.')
    os.chdir(curdir)

    # clean up
    os.remove(srcfile)
    os.remove(recfile)

    print('Prepare fmtomo dir first run: Success')


def setup_fmtomo_sim(database_path_dmt, fmtomo_dir, fname_extension, sub_phases, ncores=None, model='ak135',
                     min_picks_per_phase=10, write_vtk=True, check_notesfile=True, fname_station_blacklist=None,
                     no_write_init_nodes=False, no_recalculate_init_nodes=False, phase_types=('P, S')):
    '''
    main function of this program, writes picks and input source file for FMTOMO obsdata program
    '''

    assert os.path.isdir(database_path_dmt), 'Unrecognized directory {}'.format(database_path_dmt)
    assert os.path.isdir(fmtomo_dir), 'Unrecognized directory {}'.format(fmtomo_dir)

    tstart = datetime.now()
    print('Starting script at {}'.format(tstart))
    print('Check notesfile set to {}'.format(check_notesfile))

    # save means of event traveltimes to dictionary for statistical analysis
    #tt_means = {}

    fname_fmtomo_nodes = os.path.join(fmtomo_dir, 'init_nodes.out')

    # do first initialisation of FMTOMO to generate init_nodes file if required
    if not no_recalculate_init_nodes:
        prepare_fmtomo_dir_first_run(fmtomo_dir)

    eventdirs = glob.glob(os.path.join(database_path_dmt, '*.?'))
    nEvents = len(eventdirs)
    # create directory that will contain the picks
    picksdir = os.path.join(fmtomo_dir, 'picks')
    if not os.path.isdir(picksdir):
        os.mkdir(picksdir)
    # track and count ALL source locations & phases (P and S) for fmtomo_tools
    associations_str = ''
    nsrc_total = 0
    # iterate over P and S to create one model each
    for phase_type in phase_types:
        sourcefile_str = ''
        nsrc = 0
        # iterate over all events in "database_path_dmt"
        for eventindex, eventdir in enumerate(eventdirs):
            print('Working on {}-picks for event {} ({}/{})'.format(phase_type, eventdir,
                                                                    eventindex + 1, len(eventdirs)))
            if check_notesfile and not check_event_notes(eventdir):
                continue
            # create ObsPy event from .pkl file in dmt eventdir
            event = get_event_obspy_dmt(eventdir)
            if len(event.origins) > 1:
                raise Exception('Ambiguous origin information for event {}'.format(event))
            origin = event.origins[0]
            # get all picks from PyLoT *.xml file
            picks = get_picks(eventdir, extension=fname_extension)
            if not picks:
                print('No picks for event {} found.'.format(eventdir))
                continue
            # remove picks for blacklisted stations
            if fname_station_blacklist:
                picks, n_deleted_blacklist = remove_blacklisted_station_picks(picks, fname_station_blacklist)
            # get metadata for current event from dmt_database
            metadata = get_metadata(eventdir)
            # get a dictionary containing coordinates for all stations
            stations_dict = metadata.get_all_coordinates()
            # catch event id
            event_id = get_event_id(eventdir)
            # map specific phases to another phase (Pdiff -> P)
            merge_phases = {'Pdiff': 'P'}
            for key, value in merge_phases.items():
                print('Will map phase {} to phase {} if present.'.format(key, value))
            # assign all picks of this event to a phase and add to sorted_picks dictionary
            sorted_picks = sort_picks_by_phase(picks, stations_dict, origin,
                                               sub_phases[phase_type], model, merge_phases=merge_phases)
            sorted_picks = remove_uncommon_picks(sorted_picks, min_picks_per_phase)
            # the following should not be necessary when calculating traveltimes on borders using TauPy
            #sorted_picks = translate_phase_names_TAUP(sorted_picks)
            #
            # iterate over sorted picks and write a traveltime file for each phase type
            for phase, picks_list in sorted_picks.items():
                pickfile_base = '{}_{}'.format(event_id, phase)
                pickfile_name = pickfile_base + '.ttf'
                init_nodes_name = pickfile_base + '.nodes'
                vtk_file_name = pickfile_base + '.vtk'
                pickfile = os.path.join(picksdir, pickfile_name)
                init_nodes_file = os.path.join(picksdir, init_nodes_name)
                vtk_file = os.path.join(picksdir, vtk_file_name)
                # remove source time
                picks_list = subtract_source_time(picks_list, origin.time)
                # save mean for statistical analysis
                #tt_means[pickfile_base] = mean
                # create a list with all "true" phases used for taupy to calculate boundary ttimes/ray parameters
                phases = list(set([item['true_phase'] for item in picks_list]))
                if no_write_init_nodes == False:
                    if no_recalculate_init_nodes and os.path.isfile(init_nodes_file):
                        print('Found previously calculated init nodes. Will not recalculate.')
                    else:
                        # calculate travel times and ray parameters for boundary nodes and write to file
                        points_list = initNodesFm3d(fname_fmtomo_nodes, origin, model, phases, ncores=ncores)
                        # in case initNodes fails and returns None continue with next phase
                        if not points_list:
                            print('No points list, initNodesFm3d failed for event.')
                            continue
                        if not write_init_points(points_list, init_nodes_file):
                            print('Write init points failed for event.')
                            continue
                        if write_vtk==True:
                            write_vtk_file(points_list, vtk_file)
                # write picks to pickfile for fm3d
                write_picks_to_pickfile(pickfile, phase, picks_list, stations_dict, origin)
                # add pickfile to sourcefile string and count number of pickfiles
                sourcefile_str += '1    1    {}\n'.format(pickfile_name)
                # add source location, phaseID and .nodes filename to association string
                source_string = '{phase} {lat} {lon} {depth} {fname}\n'
                associations_str += source_string.format(phase=phase, lat=origin.latitude,
                                                         lon=origin.longitude, depth=origin.depth,
                                                         fname=init_nodes_name)
                nsrc += 1
                nsrc_total += 1
            average_time = (datetime.now() - tstart) / (eventindex + 1)
            print('Average time for event (phase {}): {}'.format(phase_type, average_time))
            print('ETA for {}-events: {}'.format(phase_type, tstart + nEvents * average_time))

        write_src_file(fmtomo_dir, phase_type, nsrc, sourcefile_str)

    write_assc_file(fmtomo_dir, nsrc_total, associations_str)
    #write_json(tt_means, os.path.join(fmtomo_dir, 'ttmeans.json'))

    print('Script finished! Good Bye!')


def write_src_file(fmtomo_dir, phase_type, nsrc, sourcefile_str):
    # write input_source_file for obsdata
    input_source_file = open(os.path.join(fmtomo_dir, 'input_source_file_{}.in'.format(phase_type)), 'w')
    input_source_file.write('{}\n'.format(nsrc))
    input_source_file.write(sourcefile_str)
    input_source_file.close()


def write_assc_file(fmtomo_dir, nsrc_total, associations_str):
    # write input_associations file for fm3d
    input_assc_file = open(os.path.join(fmtomo_dir, 'input_associations_file.in'), 'w')
    input_assc_file.write('{}\n'.format(nsrc_total))
    input_assc_file.write(associations_str)
    input_assc_file.close()


def write_json(object, fname):
    with open(fname, 'w') as outfile:
        json.dump(object, outfile, sort_keys=True, indent=4)


def write_vtk_file(points_list, fname):
    outfile = open(fname, 'w')

    nPoints = len(points_list)

    outfile.write('# vtk DataFile Version 2.0\n')
    outfile.write('FMM Init points\n')
    outfile.write('ASCII\n')
    outfile.write('DATASET POLYDATA\n')
    outfile.write('POINTS {} float\n'.format(nPoints))

    for point in points_list:
        lat = point['pt_lat']
        lon = point['pt_lon']
        rad = point['pt_R']
        x, y, z = pol2cart(lat, lon, rad)
        outfile.write('{x} {y} {z}\n'.format(x=x, y=y, z=z))

    # write number of vertices and their indices
    outfile.write('\nVERTICES {} {}\n'.format(nPoints, 2*nPoints))
    for index, point in enumerate(points_list):
        outfile.write('{} {}\n'.format(1, index))

    # write header with number of data points
    outfile.write('\nPOINT_DATA {}\n'.format(nPoints))

    # write header for traveltimes
    outfile.write('SCALARS traveltime float 1\n')
    outfile.write('LOOKUP_TABLE default\n')

    for point in points_list:
        time = point.get('time')
        outfile.write('{}\n'.format(time if time else 0))

    # write header for point indices
    outfile.write('SCALARS point_index integer 1\n')
    outfile.write('LOOKUP_TABLE default\n')

    for point in points_list:
        pt_index = point.get('pt_index')
        outfile.write('{}\n'.format(pt_index if pt_index else 0))

    outfile.write('VECTORS tt_grad float\n')
    for point in points_list:
        lat = point['pt_lat']
        lon = point['pt_lon']
        r_comp = point.get('ray_param_km_z_comp')
        n_comp = point.get('ray_param_km_n_comp')
        e_comp = point.get('ray_param_km_e_comp')
        sx, sy, sz = pol2cart_vector(lat, lon, n_comp, e_comp, r_comp)
        outfile.write('{sx} {sy} {sz}\n'.format(sx=sx if sx else 0.,
                                                sy=sy if sy else 0.,
                                                sz=sz if sz else 0.,))

    outfile.close()




def write_init_points(points_list, outfile):
    fid = open(outfile, 'w')
    print('Writing {} init points to file {}.'.format(len(points_list), outfile))
    output_str = ''
    # count number of points for file header
    npoints = 0
    for point_dict in points_list:
        nArrivals = point_dict.get('nArrivals')
        if not nArrivals:
            #print('No arrivals for point:')
            #print(point_dict)
            continue
        if nArrivals > 1:
            fid.close()
            os.remove(outfile)
            warning_template = 'Ambiguous information for point: {}, having {} different arrivals. Skipping event.'
            print(warning_template.format(point_dict, nArrivals))
            return False
        output_template = '{index} {time} {ray_param_z} {ray_param_n} {ray_param_e}\n'
        output_str += (output_template.format(index=point_dict['pt_index'],
                                              time=point_dict.get('time'),
                                              ray_param_z=point_dict.get('ray_param_km_z_comp'),
                                              ray_param_n=point_dict.get('ray_param_km_n_comp'),
                                              ray_param_e=point_dict.get('ray_param_km_e_comp')))
        npoints += 1
    fid.write('{}\n'.format(npoints))
    fid.write(output_str)
    fid.close()

    return True


def initNodesFm3d(filename_init_nodes, source_origin, model, phases, min_dist=25., ncores=None, R=6371000.):
    # This function uses obspy TauPy to calculate travel times and ray parameters on boundary points of the fm3d grid

    def exposed_sides(grid_boundaries, src_lat, src_lon):
        # check which sides of grid are exposed to source (taken from fm3d teleseismic.f90 code)
        # here: baz is calculated, not az
        north = grid_boundaries['north']
        east = grid_boundaries['east']
        south = grid_boundaries['south']
        west = grid_boundaries['west']

        # north
        baz1 = gps2dist_azimuth(north, west, src_lat, src_lon, a=R, f=0)[1]
        baz2 = gps2dist_azimuth(north, east, src_lat, src_lon, a=R, f=0)[1]
        exposed_north = all(np.cos(np.deg2rad(baz)) > 0.01 for baz in [baz1, baz2])
        # east
        baz1 = gps2dist_azimuth(north, east, src_lat, src_lon, a=R, f=0)[1]
        baz2 = gps2dist_azimuth(south, east, src_lat, src_lon, a=R, f=0)[1]
        exposed_east = all(np.sin(np.deg2rad(baz)) > 0.01 for baz in [baz1, baz2])
        # south
        baz1 = gps2dist_azimuth(south, west, src_lat, src_lon, a=R, f=0)[1]
        baz2 = gps2dist_azimuth(south, east, src_lat, src_lon, a=R, f=0)[1]
        exposed_south = all(np.cos(np.deg2rad(baz)) < -0.01 for baz in [baz1, baz2])
        # west
        baz1 = gps2dist_azimuth(north, west, src_lat, src_lon, a=R, f=0)[1]
        baz2 = gps2dist_azimuth(south, west, src_lat, src_lon, a=R, f=0)[1]
        exposed_west = all(np.sin(np.deg2rad(baz)) < -0.01 for baz in [baz1, baz2])

        exposed_dict = dict(north=exposed_north, east=exposed_east, south=exposed_south, west=exposed_west,
                            rbot=True)

        return exposed_dict

    def point_valid(pt_lat, pt_lon, pt_R, exposed_dict, grid_boundaries, R_earth):
        # check whether point has to be active or not depending on its position to the source

        def check_point(actual, desired, threshold=0.001):
            # checks receiver (boundary) point orientation by comparison to corner boundary points
            return abs(actual - desired) < threshold

        # check for negative depth (above surface)
        if pt_R > R_earth:
            return False

        # check if point belongs to bottom interface
        #if check_point(pt_R, grid_boundaries['rbot']):
        #    return True

        pt_faces_dir = {}
        pt_exposed_dir = {}

        directions = {'north': pt_lat,
                      'east': pt_lon,
                      'south': pt_lat,
                      'west': pt_lon,
                      'rbot': pt_R,}

        for direction, pt_lat_or_lon in directions.items():
            # check if point belongs to boundary of this direction and save to pt_faces_dir
            pt_direction_check = check_point(pt_lat_or_lon, grid_boundaries[direction])
            pt_faces_dir[direction] = pt_direction_check
            # check if point is exposed to source in this direction and save to pt_exposed_dir
            pt_exposed_dir[direction] = pt_direction_check and exposed_dict[direction]

        # compare number of points facing source direction to the actual direction they are facing
        return sum(pt_faces_dir.values()) == sum(pt_exposed_dir.values())


    import datetime
    now = datetime.datetime.now()
    infile = open(filename_init_nodes, 'r')
    R_earth = 6371.

    # read input files containing fm3d boundary points in each line in the order: lat lon radius
    init_nodes = infile.readlines()
    nPoints = len(init_nodes)
    # split lines and convert to floats except for first value which is converted to int
    init_nodes = [[float(val) if index != 0 else int(val) for index, val in enumerate(line.split())]
                  for line in init_nodes]

    grid_boundaries = dict(north=max(init_nodes, key=lambda x: x[1])[1],
                           east=max(init_nodes, key=lambda x: x[2])[2],
                           south=min(init_nodes, key=lambda x: x[1])[1],
                           west=min(init_nodes, key=lambda x: x[2])[2],
                           rbot=min(init_nodes, key=lambda x: x[3])[3],
                           rtop=max(init_nodes, key=lambda x: x[3])[3])

    # get source coordinates
    src_lat, src_lon, src_depth = source_origin.latitude, source_origin.longitude, source_origin.depth
    #src_lat = 0; src_lon = 180; src_depth=50

    # calculate which sides are exposed to source
    exposed_dict = exposed_sides(grid_boundaries, src_lat, src_lon)

    # iterate over all points and calculate distance. Append a dictionary to an input list for
    # multiprocessing worker for each point
    points_list = []
    for point in init_nodes:
        pt_index, pt_lat, pt_lon, pt_R = point
        pt_depth = R_earth - pt_R
        dist = locations2degrees(pt_lat, pt_lon, src_lat, src_lon)

        # check minimum distance for this point
        if dist < min_dist:
            warnings.warn('Distance for point {} less than minimum'
                          ' distance ({} km). Skipping event!'.format(point, min_dist))
            return

        # check if point is exposed to source and has to be initiated
        if not point_valid(pt_lat, pt_lon, pt_R, exposed_dict, grid_boundaries, R_earth):
            continue

        input_dict = {'pt_index': pt_index,
                      'pt_depth': pt_depth,
                      'pt_R': pt_R,
                      'pt_lat': pt_lat,
                      'pt_lon': pt_lon,
                      'src_lat': src_lat,
                      'src_lon': src_lon,
                      'src_depth': src_depth,
                      'dist': dist,
                      'model': model,
                      'phases': phases,
                      }
        points_list.append(input_dict)

    print('n Points init:', len(points_list))
    #plot_points(points_list)

    rvals = []
    pool = multiprocessing.Pool(ncores, maxtasksperchild=1000)
    for rval in pool.imap(taup_worker, points_list, chunksize=10):
        rvals.append(rval)
    pool.close()
    pool.join()

    print('Done after {}'.format(datetime.datetime.now() - now))

    #plot_points(points_list)
    return rvals


def taup_worker(input_dict):
    # initiate model for TauP method
    model = TauPyModel(model=input_dict['model'])
    huge_time = 1e7
    try:
        arrivals = model.get_ray_paths(input_dict['src_depth'], input_dict['dist'],
                                       phase_list=input_dict['phases'],
                                       receiver_depth_in_km=input_dict['pt_depth'], )
        if len(arrivals) < 1:
            #print('No arrivals for phase {}'.format(input_dict['phase']))
            #print(input_dict)
            return input_dict

        arr = arrivals[0]

        input_dict['nArrivals'] = len(arrivals)
        baz = gps2dist_azimuth(input_dict['pt_lat'], input_dict['pt_lon'],
                               input_dict['src_lat'], input_dict['src_lon'],
                               a=6371.*1e3, f=0)[1]

        # calculate traveltime gradient, division by R to transform from rad to km
        # multiply with -1 to get values for azimuth instead of back-azimuth
        ray_param_km_z_comp = arr.ray_param / np.tan(np.deg2rad(arr.incident_angle)) / input_dict['pt_R']
        ray_param_km_n_comp = arr.ray_param * (-1) * np.cos(np.deg2rad(baz)) / input_dict['pt_R']
        ray_param_km_e_comp = arr.ray_param * (-1) * np.sin(np.deg2rad(baz)) / input_dict['pt_R']
        input_dict['ray_param_km_z_comp'] = ray_param_km_z_comp
        input_dict['ray_param_km_n_comp'] = ray_param_km_n_comp
        input_dict['ray_param_km_e_comp'] = ray_param_km_e_comp

        input_dict['time'] = arr.time

    except Exception as e:
        print(e, input_dict)
    return input_dict


def plot_points(points_list):
    import matplotlib.pyplot as plt
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    x = []
    y = []
    z = []
    color = np.full(len(points_list), np.nan)
    for index, points_dict in enumerate(points_list):
        x.append(points_dict['pt_lat'])
        y.append(points_dict['pt_lon'])
        z.append(points_dict['pt_depth'])
        arrivals = points_dict.get('arrivals')
        if arrivals:
            color[index] = arrivals[0].time
    ax.scatter(x, y, z, c=color)
    plt.show()


def subtract_source_time(picks_list, origin_time):
    # WAS FUNCTION DEMEAN, but demean after calculating residuals using rtimes_tele in FMTOMO!!
    for phase_dict in picks_list:
       taup_time = phase_dict['taup_time']
       pick = phase_dict['pick']
       pick_time = pick.time
       ttres = pick_time - taup_time
       # I think doing this is wrong because taup times do NOT include station elevation!!! 5.3.20
       phase_dict['ttres'] = ttres

    mean = np.mean([phase_dict['pick'].time.timestamp for phase_dict in picks_list])

    for phase_dict in picks_list:
        phase_dict['ttres_demeaned'] = phase_dict['ttres'] - mean # see above comment
        phase_dict['abstimes'] = phase_dict['pick'].time - origin_time

    return picks_list


def translate_phase_names_TAUP(sorted_picks):
    # PKP and PKIKP are all just 'P' phases in fm3d (not very sure about this!)
    translations = {'PKIKP': 'PKP',
                    'SKIKS': 'SKS',}
    for phase_name in sorted_picks.keys():
        if phase_name in translations.keys():
            sorted_picks[translations[phase_name]] = sorted_picks.pop(phase_name)
    return sorted_picks


def write_picks_to_pickfile(pickfile, phase, picks_list, stations_dict, origin, picks_mode='abs'):
    fid = open(pickfile, 'w')
    header = '{npicks}\n' \
             '{lat_src:<15} {lon_src:<15} {depth_src:<15}\n' \
             '{phase_name}\n'
    # pickfile header including npicks, src coords, phasetype
    formatted_str = header.format(npicks=len(picks_list),
                                  lat_src=origin.latitude,
                                  lon_src=origin.longitude,
                                  depth_src=origin.depth,
                                  phase_name=phase)
    fid.write(formatted_str)

    # write picks for each station to pickfile
    for phase_dict in picks_list:
        pick = phase_dict['pick']
        seed_id = pick.waveform_id.get_seed_string()
        # travel time residual (demeaned or abs)
        if picks_mode == 'abs':
            picktime = phase_dict['abstimes']
        elif picks_mode == 'res':
            warnings.warn('Using residuals here might not be exact. See above code where ttres_demeaned is calculated.')
            picktime = phase_dict['ttres_demeaned']
        else:
            raise IOError(f'Unknown picks_mode {picks_mode}')
        uncertainty = pick.time_errors.uncertainty
        network, station = seed_id.split('.')[:2]
        # get stations coords from metadata dict
        station_coords = stations_dict.get('{}.{}'.format(network, station))
        # prepare line to be written to pickfile and format, use traveltime residual
        line = '{lat_rec:<15} {lon_rec:<15} {depth_rec:<15} {picktime:<15} {uncert:15}\n'
        formatted_line = line.format(lat_rec=station_coords['latitude'],
                                     lon_rec=station_coords['longitude'],
                                     depth_rec=station_coords['elevation'] * (-1e-3),
                                     picktime=picktime,
                                     uncert=uncertainty)
        fid.write(formatted_line)
    print('Wrote {} picks for phase {} to file {}'.format(len(picks_list), phase, pickfile))
    fid.close()


def remove_blacklisted_station_picks(picks, fname_blacklist, verbosity=1):
    blacklisted_stations = get_station_blacklist(fname_blacklist)
    deleted_picks = []
    deleted_stations = []
    for index, pick in list(reversed(list(enumerate(picks)))):
        seed_id = pick.waveform_id.get_seed_string()
        network, station = seed_id.split('.')[:2]
        nwst_id = '{}.{}'.format(network, station)
        if nwst_id in blacklisted_stations.keys():
            timewindow = blacklisted_stations[nwst_id]
            if not timewindow == 'always':
                tstart, tstop = [UTCDateTime(time) for time in timewindow.split('to')]
                if not tstart <= picks[index].time <= tstop:
                    continue
            deleted_picks.append(picks.pop(index))
            deleted_stations.append(nwst_id)
    if verbosity > 0:
        print('Deleted picks for blacklisted stations:\n{}'.format(deleted_stations))
    return picks, deleted_stations


def get_station_blacklist(fname_csv):
    with open(fname_csv, 'r') as infile:
        # skip first line
        blacklist = infile.readlines()[1:]
    blacklisted_stations = {}
    for line in blacklist:
        network, station, time = line.split(',')[:3]
        nwst_id = '{}.{}'.format(network, station)
        blacklisted_stations[nwst_id] = time
    return blacklisted_stations


def remove_uncommon_picks(sorted_picks, min_picks_per_phase):
    for phase_name, picks_list in reversed(list(sorted_picks.items())):
        if len(picks_list) < min_picks_per_phase:
            msg = 'Removed picks for phase {}, as there are only {} picks given (threshold is {})'
            print(msg.format(phase_name, len(picks_list), min_picks_per_phase))
            del(sorted_picks[phase_name])
    return sorted_picks


def sort_picks_by_phase(picks, stations_dict, source_origin, phase_types, model,
                        max_phase_diff=50., merge_phases=None, verbosity=0):
    '''
    # First: Iterate through all picks to estimate possible phases for source/location combination, then assign
    # each pick to one phase and return sorted picks dictionary, as there has to be one pickfile for each phasetype

    :param picks:
    :param stations_dict:
    :param source_origin:
    :param phase_types:
    :param model:
    :param max_phase_diff:
    :param merge_phases:  assign a phase (key) to another phase (value) e.g. {Pdiff: P}
    :param verbosity:
    :return:
    '''
    print('Starting to sort picks by phase calculating theoretical travel times for each pick...')
    phases_dict = {}
    for pick in picks:
        # PROBLEM: seed_id from PyLoT does not contain Location ID!!! Makes it hard to find station coords
        # workaround: use stations_dict (ignore channel and location ID) instead of metadata.get_coordinates()
        seed_id = pick.waveform_id.get_seed_string()
        network, station = seed_id.split('.')[:2]
        phaseID = pick.phase_hint
        uncertainty = pick.time_errors.uncertainty
        # skip different phase
        if not phaseID in phase_types:
            continue
        # pick invalid if no uncertainty is given
        if not uncertainty:
            continue
        station_coords = stations_dict.get('{}.{}'.format(network, station))
        if not station_coords:
            print('Could not find coordinates for station: {}'.format(seed_id))
            continue
        # estimate phase type by taking the closest phase from Tau-P method (time is relative to source origin)
        phase_name, phase_time_rel, phase_diff = get_closest_taup_phase(source_origin, station_coords,
                                                                        phaseID, pick.time, model)
        if phase_diff > max_phase_diff:
            if verbosity > 0:
                print ('Warning, max_diff too large (> {} s) for phase {} at {}. Skipping'.format(max_phase_diff,
                                                                                                  phase_name, seed_id))
            continue
        phase_time = source_origin.time + phase_time_rel

        # merge phase if selected, keep track of original phase for TauPy node initiation
        true_phase = phase_name
        if merge_phases and phase_name in merge_phases.keys():
            phase_name = merge_phases[phase_name]

        if not phase_name in phases_dict.keys():
            print('Adding phase to sorted picks dictionary: {}'.format(phase_name))
            phases_dict[phase_name] = []
        phases_dict[phase_name].append({'pick': pick,
                                        'taup_time': phase_time,
                                        'true_phase': true_phase,})
    return phases_dict

def get_closest_taup_phase(source_origin, station_coords, phase, pick_time, model='ak135'):
    '''
     Estimate phase that was picked by searching for earliest P/S phase arriving at
     station_coords for given source_coords using TauPy. As PyLoT only searches for first arrivals
     using the same method, this should(!) yield the correct phase.
    :return:
    '''
    phases = {'P': [],
              'S': []}
    phase_lists = {'P': ['ttp'],
                   'S': ['tts']}

    # possible phases for fm3d (*K* and *KIK* -> *)
    #phase_list = {'P': ['P', 'PKP', 'PKiKP', 'PKIKP', 'PcP', 'ScP', 'SKP', 'PKKP', 'SKKP', 'PP',],
    #              'S': ['S', 'SKS', 'SKIKS', 'ScS']}

    # in case pick phase hint is just P or S
    if phase in phases.keys():
        phase_list = phase_lists[phase]
        common_phase = phase
    # in case an explicit phase name is given use only this phase
    else:
        phase_list = [phase]
        common_phase = identifyPhaseID(phase)

    model = TauPyModel(model=model)
    arrivals = model.get_travel_times_geo(source_origin.depth,
                                          source_origin.latitude,
                                          source_origin.longitude,
                                          station_coords['latitude'],
                                          station_coords['longitude'],
                                          phase_list)

    # identifies phases from arrivals as P or S phase, not necessary when using 'ttp' or 'tts' for get_travel_times_geo
    for arr in arrivals:
        phases[identifyPhaseID(arr.phase.name)].append(arr)

    source_time = source_origin.time

    if not arrivals:
        raise Exception('No arrivals found for source.')

    # get first P or S onsets from arrivals list
    arr, min_diff = min([(arr, abs(source_time + arr.time - pick_time)) for arr in phases[common_phase]],
                        key=lambda t: t[1])
    return arr.name, arr.time, min_diff


def get_event_obspy_dmt(eventdir):
    event_pkl_file = os.path.join(eventdir, 'info', 'event.pkl')
    if not os.path.exists(event_pkl_file):
        raise IOError('Could not find event path for event: {}'.format(eventdir))
    event = qml_from_obspyDMT(event_pkl_file)
    return event


def get_event_pylot(eventdir, extension=''):
    event_id = get_event_id(eventdir)
    filename = os.path.join(eventdir, 'PyLoT_{}{}.xml'.format(event_id, extension))
    if not os.path.isfile(filename):
        return
    cat = read_events(filename)
    return cat[0]

def get_event_id(eventdir):
    event_id = os.path.split(eventdir)[-1]
    return event_id

def get_picks(eventdir, extension=''):
    event_id = get_event_id(eventdir)
    filename = 'PyLoT_{}{}.xml'
    filename = filename.format(event_id, extension)
    fpath = os.path.join(eventdir, filename)
    fpaths = glob.glob(fpath)
    if len(fpaths) == 1:
        cat = read_events(fpaths[0])
        picks = cat[0].picks
        return picks
    elif len(fpaths) == 0:
        print('get_picks: File not found: {}'.format(fpath))
        return
    print(f'WARNING: Ambiguous pick file specification. Found the following pick files {fpaths}\nFilemask: {fpath}')
    return

def init_sources_file(nsrc, filename='sources.in'):
    infile_source = open(filename, 'w')
    infile_source.write('{}\n'.format(nsrc))
    return infile_source

def init_receivers_file(nrec, filename='receivers.in'):
    infile_rec = open(filename, 'w')
    infile_rec.write('{}\n'.format(nrec))
    return infile_rec

def append_source(fid, origin):
    pass

def append_receiver(fid, station_coords):
    # TODO check if order is correct
    fid.write('{rad:15} {lat:15} {lon:15} \n'.format(rad=station_coords['elevation'],
                                                     lat=station_coords['latitude'],
                                                     lon=station_coords['longitude']))


def organize_receivers(fnin, unique=False):
    ''' Open FMTOMO receivers.in file and read position of each receiver into a dict.'''

    with open(fnin, 'r') as infile:
        nRec = int(infile.readline())
        rec_dict = {}
        for rec_number in range(1, nRec + 1):
            rad, lat, lon = [float(value) for value in infile.readline().split()]
            # dummy read next 3 lines
            for ind in range(3):
                infile.readline()

            receiver = {'rad': rad,
                        'lat': lat,
                        'lon': lon, }
            if unique:
                if receiver in rec_dict.values():
                    continue

            rec_dict[rec_number] = receiver

    return rec_dict


def organize_sources(fnin):
    ''' Open FMTOMO sources.in file and read position and phase of each source into a dict.'''
    with open(fnin, 'r') as infile:
        nSrc = int(infile.readline())
        src_dict = {}
        for src_number in range(nSrc):
            src_number += 1
            teleseismic_flag = int(infile.readline())
            if teleseismic_flag != 1: raise ValueError('Source not teleseismic!')
            phase = infile.readline().strip()
            rad, lat, lon = [float(value) for value in infile.readline().split()[:3]]
            # dummy read next 4 lines
            for ind in range(4):
                infile.readline()
            src_dict[src_number] = {'phase': phase,
                                    'depth': rad,
                                    'lat': lat,
                                    'lon': lon,}
    return src_dict


def organize_event_names(fnin):
    infile = open(fnin, 'r')
    events = [line.split()[-1].strip() for line in infile.readlines()[1:]]
    infile.close()
    return events


def export_otimes(otimes, fn_out):
    np.savetxt(fn_out, otimes, fmt=['%8d', '%8d', '%8d', '%8d', '% 4.6f', '% 4.6f'],
               header=str(len(otimes)), comments='')
    print('Wrote file:', fn_out)


if __name__ == '__main__':
    # testing area
    prepare_fmtomo_dir_first_run('/data/AdriaArray_Data/fmtomo_adriaarray/v0/test_init')