#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Different functions to modify FMTOMO velocity(inversion) grid
import argparse
import numpy as np
from math import erf

from pylot.tomography.fmtomo_tools.fmtomo_grid_utils import read_vgrid, write_vgrid, write_vtk


def main(fname_in, fname_out, fname_vtk_out=None, cov_perc=0.15, bounds=[40, 60]):
    vgrid, gridN, gridDelta, gridStart = read_vgrid(fname_in)

    print('Modifying covariances...')
    vgrid['covs'] = []
    for depth, vp in zip(vgrid['depths'], vgrid['vps']):
        vgrid['covs'].append(covariance_for_depth(depth, vp, cov_perc, bounds))

    write_vgrid(vgrid, npts=gridN, delta=gridDelta, start=gridStart, fname=fname_out)
    if fname_vtk_out:
        write_vtk(vgrid, fname_vtk_out, write_data=['vps', 'covs'])


def modify_vgrid_box(vgrid, vgrid_key, min_lon, max_lon, min_lat, max_lat, min_depth, max_depth,
                     val_center, val_border, scale_factor, use_erf=True, extend_to_top=False):
    '''
    Modify a value (vgrid_key) of vgrids.in using an error function. Value will be increase smoothly to the borders
    of the external model. This function was used to prevent changes of the inversion result in the region of the
    initial crustal model that was corrected for. Value smoothly increases at the borders of the initial model.
    '''

    # get central points of external model (here value will be minimal)
    c_lon = 0.5 * (min_lon + max_lon)
    c_lat = 0.5 * (min_lat + max_lat)
    c_depth = 0.5 * (min_depth + max_depth)

    # get half the extent of the model dimensions for function scaling
    dlon = 0.5 * (max_lon - min_lon)
    dlat = 0.5 * (max_lat - min_lat)
    ddepth = 0.5 * (max_depth - min_depth)

    print('Modifying {}...'.format(vgrid_key))
    print('Center {key}: {val_c}, border {key}: {val_b}'.format(key=vgrid_key, val_c=val_center, val_b=val_border))
    if not vgrid_key in vgrid.keys() or not vgrid[vgrid_key]:
        vgrid[vgrid_key] = list(np.ones(len(vgrid['depths'])))
    for index, tup in enumerate(zip(vgrid['lons'], vgrid['lats'], vgrid['depths'])):
        lon, lat, depth = tup
        if (min_lon <= lon <= max_lon and min_lat <= lat <= max_lat and depth <= max_depth):
            x = abs(lon - c_lon) / dlon
            y = abs(lat - c_lat) / dlat
            if extend_to_top and depth < c_depth:
                z = 0
            else:
                z = abs(depth - c_depth) / ddepth
            if use_erf:
                point_val = smoothing_erf(x, y, z, val_center, val_border, scale_factor=scale_factor)
            else:
                point_val = val_center
        else:
            point_val = val_border
        vgrid[vgrid_key][index] *= point_val

    return vgrid


def modify_vgrid_gradient(vgrid, vgrid_key, val_top, val_bot):
    '''
    Modify a value (vgrid_key) of vgrids.in using a linear gradient from val_top to val_bot
    '''

    print('Modifying {} with linear gradient...'.format(vgrid_key))
    print('Top: {}, bot: {}'.format(val_top, val_bot))
    if not vgrid_key in vgrid.keys():
        vgrid[vgrid_key] = list(np.ones(len(vgrid['depths'])))
    depth_min = min(vgrid['depths'])
    depth_max = max(vgrid['depths'])
    for index, tup in enumerate(zip(vgrid['lons'], vgrid['lats'], vgrid['depths'])):
        lon, lat, depth = tup
        vgrid[vgrid_key][index] *= linear_gradient(depth, depth_min=depth_min, depth_max=depth_max, val_top=val_top,
                                                   val_bot=val_bot)
    return vgrid


def linear_gradient(depth, depth_min, depth_max, val_top, val_bot):
    return (val_bot - val_top) * depth / (depth_max - depth_min) + val_top


def smoothing_erf(x, y, z, val_border, val_center, scale_factor=1.):
    a = val_center
    b = val_border
    sx = sy = sz = scale_factor
    f = (a - b) * (1. / 3. * (erf(2. * sx * (x - 1)) + erf(2. * sy * (y - 1)) + erf(2. * sz * (z - 1))) + 1.) + b
    return f


def covariance_for_depth(depth, vp, cov_perc, bounds):
    '''
    Function that returns covariance values for certain depths, written to give low variance to crust, intermediate
    to an interlayer and high variance to everything else.
    :param depth: depth in kilometers
    :return: covariance
    '''
    if depth <= bounds[0]:
        return 0.05
    elif bounds[0] < depth <= bounds[1]:
        return 0.5 * cov_perc * vp
    else:
        return cov_perc * vp

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Prepare grid for fm3d teleseismic hybrid calculation.')
    parser.add_argument('fname_in', help='input filename (vgrids.in)')
    parser.add_argument('fname_out', help='output filename (vgrids_new.in)')
    parser.add_argument('--fname_out_vtk', default=None, help='vtk output filename')
    parser.add_argument('--cov_border', default=1.0, help='covariance for the model outside crustal model boundaries')
    parser.add_argument('--cov_center', default=0.05, help='maximum covariance in the center of the crustal model')
    parser.add_argument('--smooth_border', default=1., help='smoothing factor for the rest of the box')
    parser.add_argument('--smooth_center', default=0.5, help='smoothing factor inside SWATH-D box')

    args = parser.parse_args()

    vgrid, gridN, gridDelta, gridStart = read_vgrid(args.fname_in)

    #main(args.fname_in, args.fname_out, args.fname_out_vtk)

    #vgrid = modify_vgrid_gradient(vgrid, 'smoothfactors', 1.0, 2.0)

    # bounds for waldhauser residual corrections
    wh_bounds = dict(lon0 = 3.23, lon1 = 13.96, lat0 = 43.18, lat1 = 49.45)
    vgrid = modify_vgrid_box(vgrid, vgrid_key='covs', min_lon=wh_bounds['lon0'], max_lon=wh_bounds['lon1'],
                             min_lat=wh_bounds['lat0'], max_lat=wh_bounds['lat1'], min_depth=-5.0, max_depth=60.0,
                             val_center=float(args.cov_center), val_border=float(args.cov_border),
                             scale_factor=3, use_erf=True, extend_to_top=True)

    #vgrid = modify_vgrid_box(vgrid, vgrid_key='covs', min_lon=2.2547, max_lon=17.0999, min_lat=41.0517,
    #                         max_lat=50.4984, min_depth=-5.0, max_depth=70.0, val_center=float(args.cov_center),
    #                         val_border=float(args.cov_border), scale_factor=3, use_erf=True, extend_to_top=True)

    #vgrid = modify_vgrid_box(vgrid, vgrid_key='smoothfactors', min_lon=8.9, max_lon=15.3, min_lat=45.,
    #                         max_lat=48., min_depth=-15.0, max_depth=610.0, val_center=float(args.smooth_center),
    #                         val_border=float(args.smooth_border), scale_factor=1., use_erf=True)

    # MP MP TEST TEST TEST ++++++++++++++++++++++++++++++++++++++++++++++++
    #grid = modify_vgrid_box(grid, vgrid_key='smoothfactors', min_lon=11., max_lon=35., min_lat=30.,
    #                         max_lat=61., min_depth=-5.0, max_depth=600.0, val_center=float(args.smooth_center),
    #                         val_border=float(args.smooth_border), scale_factor=.4)
    # MP MP TEST TEST TEST ------------------------------------------

    write_vgrid(vgrid, npts=gridN, delta=gridDelta, start=gridStart, fname=args.fname_out)
    if args.fname_out_vtk:
        write_vtk(vgrid, args.fname_out_vtk, write_data=['vps', 'covs', 'smoothfactors'])


