#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import argparse
import json

import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import griddata

from obspy.geodetics.base import gps2dist_azimuth

import cartopy.crs as ccrs

from pylot.tomography.fmtomo_tools.compare_arrivals_hybrid import organize_receivers, organize_sources, organize_event_names
from pylot.tomography.map_utils import make_map
from pylot.tomography.utils import normed_figure


def plot_map(otimes_fname, sources_fname, receivers_fname, isf, rtimes_fname=None, arrivals_fname=None, colorbar=True,
             title=True, stack=False, source_ids=(), savefig_dir='', demean=True, absolute=False, clat=42.25, clon=17.5,#clat=46., clon=11.,
             file_ext='png', max_abs=3., relative_to_otimes=False, only_otimes=False):
    if savefig_dir:
        if not os.path.isdir(savefig_dir):
            os.mkdir(savefig_dir)

    src_dict = organize_sources(sources_fname)
    rec_dict = organize_receivers(receivers_fname)
    eventnames = organize_event_names(isf)

    residuals_dict = read_fmtomo_tt_file(otimes_fname, rec_dict, rtimes_fname=rtimes_fname, synth_fname=arrivals_fname,
                                         demean=demean, absolute=absolute, height_correction_vp=5.5,
                                         relative_to_otimes=relative_to_otimes, only_otimes=only_otimes)

    if stack:
        residuals_dict = stack_sources(residuals_dict)
        with open('stacked_residuals.json', 'w') as outfile:
            json.dump(residuals_dict, outfile)

    count = 0
    if savefig_dir:
        fig = normed_figure(width_cm=12, ratio=12./9.)
        #fig = plt.figure(figsize=(12, 9))
    else:
        fig = plt.figure()
    if stack or source_ids is not [] or savefig_dir:
        count = 1

    # increase point size if not stacked
    sizefactor_stacked = {True: 30,
                          False: .2e4}

    for src_id, dic in residuals_dict.items():
        eventname = eventnames[src_id - 1]
        if source_ids != []:
            if not src_id in source_ids:
                continue
        if max_abs == 'auto':
            max_abs = np.max(np.abs(np.array(dic['ttimes'])))
        # if not stack and not savefig_dir and source_ids == []:
        #     count += 1
        #     if count == 1:
        #         ax = fig.add_subplot(3, 3, count)
        #         ax0 = ax
        #     else:
        #         ax = fig.add_subplot(3, 3, count, sharex=ax0, sharey=ax0)
        #
        #     ax.text(0.1, 0.9, 'Source ID {}, {}'.format(src_id, src_dict[src_id]['phase']), transform=ax.transAxes)
        # else:
        #     ax = fig.add_subplot(111)
        if stack and title:
            plt.title('Stacked {} sources. Size relates to number of stacks per receiver.'.format(len(src_dict)),
                      y=1.05)
        ax = make_map(draw_model=False, draw_faults=True, model_legends=False, clon=clon, clat=clat,
                      width = 30, height = 21,)  # , width=6e6, height=5e6)
        if not stack:
            slat, slon = src_dict[src_id]['lat'], src_dict[src_id]['lon']
            dist = plot_source(slat, slon)
            baz = gps2dist_azimuth(clat, clon, slat, slon, a=6.371e6, f=0)[1]
            if title:
                plt.title(
                    'Plot of source {}. Distance: {:.0f}$^\circ$. BAZ: {:.0f}$^\circ$'.format(eventname, dist, baz),
                    y=1.05)

        #ax.set_xlabel('Longitude [$^\circ$]')
        #ax.set_ylabel('Latitude [$^\circ$]')

        # prepare grids for contour plot
        #lonaxis = np.linspace(min(dic['lons']), max(dic['lons']), 250)
        #lataxis = np.linspace(min(dic['lats']), max(dic['lats']), 250)
        #longrid, latgrid = np.meshgrid(lonaxis, lataxis)
        #ttimes_grid = griddata((dic['lats'], dic['lons']), dic['ttimes'], (latgrid, longrid), method='linear')
        #levels = np.linspace(min(dic['ttimes']), max(dic['ttimes']), 75)

        cmap = plt.get_cmap('seismic') if not absolute else plt.get_cmap('viridis')

        vmin = -max_abs if not absolute else None
        vmax = max_abs if not absolute else None
        sc = ax.scatter(dic['lons'], dic['lats'], c=np.array(dic['ttimes']), cmap=cmap,
                        vmin=vmin, vmax=vmax, edgecolors='grey', linewidths=0.2,
                        s=sizefactor_stacked[stack]*np.array(dic['nttimes'])/len(src_dict), zorder=10,
                        transform=ccrs.PlateCarree(), alpha=0.75)
        sc.set_edgecolor('0.')
        if colorbar:
            cb = plt.colorbar(sc, label='ttime [s]', shrink=0.5)

        if stack:
            savefig_path = os.path.join(savefig_dir, 'stacked_events.{}'.format(file_ext))
        else:
            #fname = f'baz{baz:03.0f}_dist{dist:03.0f}_{eventname}_srcID{src_id}.{file_ext}'
            fname = f'{eventname}_srcID{src_id}.{file_ext}'
            savefig_path = os.path.join(savefig_dir, fname)


        if not savefig_dir and not count % 9:
            plt.show()
            fig = plt.figure()#figsize=(16, 9), dpi=300.)
            count = 0
        if savefig_dir:
            ax.figure.savefig(savefig_path, dpi=300., bbox_inches='tight', pad_inches=0.)
            print('Wrote file {}'.format(savefig_path))
            plt.clf()

    if not savefig_dir:
        plt.show()


def plot_source(lat, lon, clat=46., clon=11.):
    #basemap.drawgreatcircle(lon, lat, clon, clat, color='k', zorder=15, linestyle='dashed')
    dist, azim, bazim = gps2dist_azimuth(lat, lon, clat, clon, a=6.371e6, f=0)

    dist_deg = dist/1000./np.pi/2./6371.*360.
    return dist_deg

    #x = np.cos(np.deg2rad(azim))
    #y = np.sin(np.deg2rad(azim))
    #print(x, y)
    #ax.plot([0, x], [0, y], 'r', zorder=15)


def read_fmtomo_tt_file(otimes_fname, rec_dict, synth_fname=None, rtimes_fname=None, demean=True, absolute=False,
                        height_correction_vp=None, R=6371., relative_to_otimes=True, only_otimes=False):
    #with open(otimes_fname, 'r') as infile:
    #    # skip first line for otimes.dat. Frist line contains N_rays
    #    obsarray = infile.readlines()[1:]
    obsarray = np.loadtxt(otimes_fname, skiprows=1)

    if height_correction_vp:
        if absolute:
            print('APPLICATION OF HEIGHT CORRECTION FOR ABS NOT IMPLEMENTED YET. Not needed!?')
            #print('Applying height correction of {} km/s'.format(height_correction_vp))
        else:
            height_correction_vp = None
            print('Will not apply height correction for relative values.')

    if synth_fname:
        synth_array = np.genfromtxt(synth_fname)

    if rtimes_fname:
        ref_array = np.genfromtxt(rtimes_fname)

    if synth_fname and rtimes_fname:
        #with open(synth_fname, 'r') as infile:
        #    synth_array = infile.readlines()
        #with open(rtimes_fname, 'r') as infile:
        #    ref_array = infile.readlines()

        #mean_obs = np.mean(obsarray[:, 4])
        #mean_synth = np.mean(synth_array[:, 4])
        #mean_ref = np.mean(ref_array[:, 4])

        #mean_rel_synth = mean_synth - mean_ref
        #print('Means:\nobserved: {}, relative {}'.format(mean_obs, mean_rel_synth))
        synth_src_means = {}

        nsrc = obsarray[-1, 1]
        for index in range(int(nsrc)):
            srcid = index + 1
            indices = np.where(synth_array[:, 1].astype(int) == srcid)
            synth_src_means[srcid] = np.mean(synth_array[indices, 4] - ref_array[indices, 4])

        print('Average mean for sources: {} s'.format(np.mean(list(synth_src_means.values()))))

        if demean:
            print('Removing mean from synthetic times...')
        else:
            print('Will not mean-correct travel times')
        #    mean_rel_synth = 0.

    residuals_dict = {}

    for index, line in enumerate(obsarray):
        rec_id = int(line[0])
        src_id = int(line[1])
        ttime = line[4] if relative_to_otimes or only_otimes else 0

        # get residuals from reference file if fname is given
        if synth_fname:
            # get synthetic time
            ttime_syn = synth_array[index][4]
        if rtimes_fname:
            # get reference time (1d travel time)
            ttime_ref = ref_array[index][4]
        if synth_fname and rtimes_fname:
            # calculate synthetic time relative to 1d travel time
            ttime_rel_syn = ttime_syn - ttime_ref
            #print(ttime_ref, ttime_syn, ttime_rel_syn, ttime, ttime-ttime_rel_syn)
            # calculate difference between relative observed and model times, multiply with -1 to get tsynth - tobs
            mean_rel_src = synth_src_means[srcid] if demean else 0.
            if not only_otimes:
                ttime = -1 * (ttime_rel_syn - ttime - mean_rel_src)
        if absolute:
            ttime = ttime_syn
        try:
            uncertainty = line[5]
        except:
            uncertainty = 0

        # create dictionary for source if not exists
        if not src_id in residuals_dict.keys():
            residuals_dict[src_id] = {'lats': [],
                                      'lons': [],
                                      'rads': [],
                                      'ttimes':[],
                                      'nttimes': [],
                                      'uncerts': []}
        # chose correct dictionary
        dic = residuals_dict[src_id]
        # get coordinates from organized receivers dictionary
        dic['lats'].append(rec_dict[rec_id]['lat'])
        dic['lons'].append(rec_dict[rec_id]['lon'])
        dic['rads'].append(rec_dict[rec_id]['rad'])
        # rad is depth?
        #elev = - rec_dict[rec_id]['rad']
        #height_correction_vp = 0
        #station_height_corr = elev / (height_correction_vp) if height_correction_vp else 0.
        #if elev < 0: print(elev, station_height_corr)
        # MP MP NO HEIGHT CORRECTION FOR FMTOMO! REFTIMES CONTAIN STATION HIGHT!!!!!
        dic['ttimes'].append(ttime) # - station_height_corr)
        # number of ttimes will be set to 1 (determines size in scatter, only relevant for stacked ttimes)
        dic['nttimes'].append(1)
        dic['uncerts'].append(uncertainty)

    return residuals_dict


def stack_sources(residuals_dict):
    all_ttimes = {}
    all_uncs = {}
    for src_id, dic in residuals_dict.items():
        for index in range(len(dic['lats'])):
            lat = dic['lats'][index]
            lon = dic['lons'][index]
            rad = dic['rads'][index]
            source_tuple = (lat, lon, rad)
            if not source_tuple in all_ttimes.keys():
                all_ttimes[source_tuple] = []
                all_uncs[source_tuple] = []
            all_ttimes[source_tuple].append(dic['ttimes'][index])
            all_uncs[source_tuple].append(dic['uncerts'][index])
    # create new dictionary in the shape of residuals dict with only one source
    stacked_dict = {}
    stacked_dict[1] = {'lats': [],
                       'lons': [],
                       'rads': [],
                       'ttimes':[],
                       'nttimes': [],
                       'misfits': []}


    for source_tuple in all_ttimes.keys():
        ttimes_list = all_ttimes[source_tuple]
        uncs_list = all_uncs[source_tuple]
        misfit = np.sum((np.array(ttimes_list) / np.array(uncs_list))**2) / len(ttimes_list)
        lat, lon, rad = source_tuple
        stacked_dict[1]['lats'].append(lat)
        stacked_dict[1]['lons'].append(lon)
        stacked_dict[1]['rads'].append(rad)
        stacked_dict[1]['ttimes'].append(np.sum(np.mean(ttimes_list)))
        stacked_dict[1]['nttimes'].append(len(ttimes_list))
        stacked_dict[1]['misfits'].append(misfit)
        #stacked_dict[1]['ttimes_list'].append(ttimes_list)

    ttimes = [d['ttimes'] for d in stacked_dict.values()]
    plt.hist(ttimes, 100)
    plt.axvline(np.mean(ttimes), c='r')
    plt.xlabel('ttime [s]')

    return stacked_dict


def compare_residuals():
    wdir = '/rscratch/minos13/marcel/fmtomo_alparray/'
    os.chdir(wdir)
    with open('alparray_mantle_diehl_crust_included_v3_hf/stacked_residuals.json', 'r') as infile:
        sr = json.load(infile)
    with open('alparray_mantle_diehl_crust_included_v3_no_density_hf/stacked_residuals.json', 'r') as infile:
        srnd = json.load(infile)
    mfdiff = np.array(sr['1']['misfits']) - np.array(srnd['1']['misfits'])
    np.mean(mfdiff)
    sc = plt.scatter(sr['1']['lons'], sr['1']['lats'], c=mfdiff, cmap='seismic', vmin=-7.5, vmax=7.5)
    cb = plt.colorbar(sc)
    title = 'Difference in station misfit after 12 iterations (density kernel - no density).' \
            ' Blue: decrease in residuals using density. Mean: {:.4f}'
    plt.title(title.format(np.mean(mfdiff)))
    plt.xlabel('Latitude (deg)')
    plt.ylabel('Longitude (deg)')
    cb.set_label('Misfit')
    sc.set_linewidth(0.2)
    sc.set_edgecolor('k')
    plt.show()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Plot residuals of fm3d output file (arrivals.dat format) on map.')
    parser.add_argument('--otimes', default='otimes.dat', help='otimes.dat file (default: otimes.dat)')
    parser.add_argument('--sources', default='sources.in', help='sources.in file (default: sources.in)')
    parser.add_argument('--receivers', default='receivers.in', help='receivers.in file (default receivers.in)')
    parser.add_argument('--sourcefile', default='input_source_file.in', help='input_source_file (default: input_source_file.in)')
    parser.add_argument('--arrivals', default=None, help='arrivals_file')
    parser.add_argument('--rtimes', default=None, help='reference_file')
    parser.add_argument('--fext', default='png', help='file extension for images (if -w)')
    parser.add_argument('-s', '--stack', action='store_true', dest='stack', help='stack picks')
    parser.add_argument('-w', '--write', action='store_true', default=False, help='write figures to disk')
    parser.add_argument('-a', '--abs', action='store_true', dest='abs', default=False, help='compute absolute values')
    parser.add_argument('-nd', '--no_demean', action='store_true', default=False, help='do not demean travel times')
    parser.add_argument('-nc', '--no_colorbar', action='store_true', default=False, help='do not plot colorbar')
    parser.add_argument('-nt', '--no_title', action='store_true', default=False, help='do not plot title')
    parser.add_argument('-no', '--no_otimes', action='store_true', default=False, help='do not plot relative to otimes')
    parser.add_argument('-oo', '--only_otimes', action='store_true', default=False, help='only plot observed times (otimes)')
    parser.add_argument('-i', '--ids', dest='source_ids', type=int, default=[], nargs='*',
                        help='Plot only one source with FMTOMO internal id(s).')

    args = parser.parse_args()

    savefig_dir = 'residual_maps_out' if args.write else ''
    plot_map(args.otimes, args.sources, args.receivers, args.sourcefile, arrivals_fname=args.arrivals,
             rtimes_fname=args.rtimes, stack=args.stack, source_ids=args.source_ids, colorbar=not args.no_colorbar,
             title=not args.no_title, savefig_dir=savefig_dir, demean=not args.no_demean, absolute=args.abs,
             file_ext=args.fext, relative_to_otimes=not args.no_otimes, only_otimes= args.only_otimes)