#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
import glob
import numpy as np
import os
import json
import matplotlib.pyplot as plt

from obspy import Catalog, read_events

from pylot.core.util.dataprocessing import Metadata


def get_metadata(eventdir):
    metadata_path = os.path.join(eventdir, 'resp')
    metadata = Metadata(inventory=metadata_path, verbosity=0)
    return metadata



def set_rc_params(textsize=7.):
    plt.style.use('/home/marcel/solid_earth.mplstyle')
    #plt.rcParams.update({'font.size': textsize,
    #                     'font.family': 'sans-serif'})


def normed_figure_ratio_width(width_cm, ratio):
    width_inch = width_cm / 2.54
    height_inch = width_inch / ratio
    return width_inch, height_inch


def normed_figure_ratio_height(height_cm, ratio):
    height_inch = height_cm / 2.54
    width_inch = height_inch * ratio
    return width_inch, height_inch


def normed_figure(width_cm=None, ratio=1.777):
    #assert ((width_cm and not height_cm) or (height_cm and not width_cm)), 'Choose either of width or height!'
    set_rc_params()
    if width_cm:
        fig = plt.figure(figsize=normed_figure_ratio_width(width_cm, ratio))
        return fig
    #elif height_cm:
    #    fig = plt.figure(figsize=normed_figure_ratio_height(height_cm, ratio))
    return


def pol2cart(lat, lon, r):
    x = r * np.cos(np.deg2rad(lat)) * np.cos(np.deg2rad(lon))
    y = r * np.cos(np.deg2rad(lat)) * np.sin(np.deg2rad(lon))
    z = r * np.sin(np.deg2rad(lat))
    return x, y, z


def cart2pol(x, y, z):
    r = np.sqrt(x**2 + y**2 + z**2)
    theta = np.rad2deg(np.arccos(z/r))
    phi = np.rad2deg(np.arctan2(y, x))
    lat = 90. - theta
    lon = phi
    return lat, lon, r


def pol2cart_vector(lat, lon, north, east, r_comp):
    if any(val is None for val in [north, east, r_comp]):
        return None, None, None
    phi = np.deg2rad(lon)
    # change north components to common spherical coordinate convention
    theta = np.deg2rad(90. - lat)
    north *= -1
    x = (np.sin(theta) * np.cos(phi) * r_comp +
         np.cos(theta) * np.cos(phi) * north -
         np.sin(phi) * east)
    y = (np.sin(theta) * np.sin(phi) * r_comp +
         np.cos(theta) * np.sin(phi) * north +
         np.cos(phi) * east)
    z = (np.cos(theta) * r_comp -
         np.sin(theta) * north)
    return x, y, z


def read_cat_obspy_dmt_database(databasedir, filemask):
    infiles = glob.glob(os.path.join(databasedir, '*.a', filemask))
    cat = Catalog()
    nPicks = 0
    for index, infile in enumerate(infiles):
        print(f'Working on: {infile} ({index + 1}/{len(infiles)})')
        event = read_events(infile)[0]
        nPicks += len(event.picks)
        cat += event

    nEvents = len(cat)
    print('Number of events: {} (filemask: {})'.format(nEvents, filemask))
    print('Total # picks: {} ({:.2f} per event)'.format(nPicks, float(nPicks)/nEvents))
    return cat


def get_event(cat, eventid):
    for event in cat.events:
        if event.resource_id.id.split('/')[-1] == eventid:
            return event


def get_pick4station(picks, network_code, station_code, method='auto'):
    for pick in picks:
        if pick.waveform_id.network_code == network_code:
            if pick.waveform_id.station_code == station_code:
                if pick.method_id.id.endswith(method):
                    return pick


def delete_picks(picks, nwst_ids_delete):
    ''' Delete picks from list in picks containing ObsPy pick objects'''
    for index, pick in list(reversed(list(enumerate(picks)))):
        seed_id = pick.waveform_id.get_seed_string()
        network, station = seed_id.split('.')[:2]
        nwst_id = '{}.{}'.format(network, station)
        if nwst_id in nwst_ids_delete:
            picks.pop(index)
            print('Removed pick: ', nwst_id)
    return picks


def save_all_station_coordinates_dmt_database(dmt_database, fn_out):
    '''
    Get all station coordinates from dmt_database and write them (unique) to json outputfile
    :param dmt_database:
    :param fn_out:
    :return:
    '''
    stations_dict = {}
    eventdirs = glob.glob(os.path.join(dmt_database, '*.?'))
    nEvents = len(eventdirs)
    for index, eventdir in enumerate(eventdirs):
        print('Working on event {} ({}/{})'.format(eventdir, index+1, nEvents))
        metadata = get_metadata(eventdir)
        current_stations_dict = metadata.get_all_coordinates()
        for nwst_id, coords in current_stations_dict.items():
            if not nwst_id in stations_dict.keys():
                stations_dict[nwst_id] = coords

    with open(fn_out, 'w') as outfile:
        json.dump(stations_dict, outfile)


def get_metadata(eventdir):
    metadata_path = os.path.join(eventdir, 'resp')
    metadata = Metadata(inventory=metadata_path, verbosity=0)
    return metadata


def get_coordinate_from_dist_baz(station_tmp, dist, baz, mode='deg'):
    ''' function copied from Andre'''
    # station_tmp: [lon, lat]
    if mode!='deg' and mode!='rad':
        print('mode hast to be ether deg or rad!')
        return None
    else:
        station = np.deg2rad(station_tmp)
        epi_tmp=[0.,0.]
        if mode=='deg':
            dist = np.deg2rad(dist)
            baz = np.deg2rad(baz)
        az = baz - np.pi
        if az < 0:
            az += 2 * np.pi
        epi_tmp[1] = np.arcsin(np.sin(station[1]) * np.cos(dist) - np.cos(station[1]) * np.sin(dist) * np.cos(az))
        if (np.cos(dist) - np.sin(station[1]) * np.sin(epi_tmp[1])) / (np.cos(station[1]) * np.cos(epi_tmp[1])) >= 0.:
            epi_tmp[0] = station[0] - np.arcsin(np.sin(dist) * np.sin(az) / np.cos(epi_tmp[1]))
        else:
            epi_tmp[0] = station[0] - np.pi + np.arcsin(np.sin(dist) * np.sin(az) / np.cos(epi_tmp[1]))
        epi=np.rad2deg(epi_tmp)
    return epi